'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

function notary() {
  var typeClasses = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};


  var processedTypeClasses = Object.keys(typeClasses).map(function (tcn) {
    return [tcn, typeClasses[tcn]];
  }).reduce(function (ptc, _ref) {
    var _ref2 = _slicedToArray(_ref, 2),
        name = _ref2[0],
        behavior = _ref2[1];

    ptc[name] = typeof behavior === 'function' ? behavior : objectToPredicate(behavior);

    return ptc;
  }, {});

  function objectToPredicate(a) {
    return function (b) {
      return Object.keys(a).every(function (k) {
        return k in b;
      });
    };
  }

  return signWithClasses.bind(undefined, processedTypeClasses);
}

function signWithClasses(typeClasses, signature, fn) {
  var _parseCheckSignature = parseCheckSignature(signature, typeClasses),
      constraints = _parseCheckSignature.constraints,
      types = _parseCheckSignature.types;

  var typedFn = function typedFn() {
    try {
      for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
        args[_key] = arguments[_key];
      }

      checkSignature(typeClasses, constraints, types.slice(0, -1), args);
      var result = fn.apply(undefined, args);
      checkSignature(typeClasses, constraints, types, args.concat([result]));
      return result;
    } catch (e) {
      e.message = 'Failed checking signature \'' + signature + '\': ' + e.message;
      throw e;
    }
  };

  return typedFn;
}

exports.notary = notary;
var sign = exports.sign = signWithClasses.bind(undefined, {});

// ----------------------------------------------------------------------------
// Parsing.
// ----------------------------------------------------------------------------

// Signature = 'Num a, Ord a, Num b => string -> ...'
// { constraints: { a: [Num, Ord], b: [Num], ... }, types: ['string', ...] }
function parseCheckSignature(signature, typeClasses) {
  var _ref3 = function () {
    try {
      var _parseConstraints = parseConstraints(signature),
          _constraints = _parseConstraints.constraints,
          unparsedTypes = _parseConstraints.unparsedTypes;

      var _types = parseTypes(unparsedTypes);
      return { constraints: _constraints, types: _types };
    } catch (e) {
      e.message = 'Malformed signature \'' + signature + '\': ' + e.message;
      throw e;
    }
  }(),
      constraints = _ref3.constraints,
      types = _ref3.types;

  // Make sure all type classes have been declared.


  Object.keys(constraints).map(function (k) {
    return constraints[k];
  })
  // Flatten array.
  .reduce(function (allNames, thisNames) {
    return allNames.concat(thisNames);
  }, []).forEach(function (typeClassName) {
    if (!typeClasses.hasOwnProperty(typeClassName)) {
      throw ReferenceError('Type class \'' + typeClassName + '\' is not defined.');
    }
  });

  return { constraints: constraints, types: types };
}

// 'Num a, Ord a, Num b => a->a->b' -> {
//    constraints: { a: ['Num', 'Ord'], b : ['Num'] },
//    unparsedTypes: ' a->a->b'
// }
function parseConstraints(signature) {
  // Omit if no constraint block is found.
  if (!/=>/.test(signature)) return { constraints: {}, unparsedTypes: signature };

  var _signature$split$redu = signature.split('').reduce(function (_ref4, char) {
    var _ref5 = _slicedToArray(_ref4, 3),
        buffer = _ref5[0],
        constraints = _ref5[1],
        finished = _ref5[2];

    if (finished) return [buffer + char, constraints, true];

    switch (char) {
      case ',':
      case '=':
        {
          // Buffer must contain class-variable pair. Parse it and continue.
          var matches = buffer.trim().match(/^(?:,\s*)?(\w+)\s+(\w+)$/);

          if (matches === null) {
            throw SyntaxError('Cannot parse class constraint \'' + buffer + '\'.');
          }

          var _matches = _slicedToArray(matches, 3),
              typeClassName = _matches[1],
              typeVariableName = _matches[2];

          constraints[typeVariableName] = constraints[typeVariableName] || [];
          constraints[typeVariableName].push(typeClassName);
          return [char, constraints, false];
        }
      case '>':
        {
          // Constraint block finished. Make sure arrow is not malformed and
          // set flag to skip the following iterations.
          if (buffer !== '=') {
            throw SyntaxError('Found extraneous \'>\' in class constraints.');
          }

          return ['', constraints, true];
        }
      default:
        {
          return [buffer + char, constraints, false];
        }
    }
  }, ['', {}, false]),
      _signature$split$redu2 = _slicedToArray(_signature$split$redu, 2),
      unparsedTypes = _signature$split$redu2[0],
      constraints = _signature$split$redu2[1];

  return { constraints: constraints, unparsedTypes: unparsedTypes };
}

function parseTypes(unparsedTypes) {
  var splittedUnparsedTypes = unparsedTypes.replace(/\s/g, '').split('->');

  if (splittedUnparsedTypes.length < 2) {
    throw new SyntaxError('Type list too short. Expected at least two types.');
  }

  return splittedUnparsedTypes.filter(function (type) {
    return type !== '()';
  }).map(function (ut) {
    var matches = ut.match(/^\[*([\w]*)\]*$/);
    // todo: do this check in syntax checking.
    if (matches === null) {
      throw new SyntaxError('Cannot parse type \'' + ut + '\'.');
    }

    var _matches2 = _slicedToArray(matches, 2),
        baseType = _matches2[1];

    var depthLeft = ut.split('[').length - 1;
    var depthRight = ut.split(']').length - 1;

    if (depthLeft !== depthRight) {
      throw new SyntaxError('Unbalanced brackets in type \'' + ut + '\'.');
    }

    return createType(baseType, depthLeft);
  });
}

// ----------------------------------------------------------------------------
// Checking.
// ----------------------------------------------------------------------------

var CONCRETE_TYPES = ['undefined', // Matches anything. Used when type can't be infered.
'string', 'number', 'boolean', 'object', 'function', 'symbol'];

function checkSignature(typeClasses, constraints, expectedTypes, values) {

  var actualTypes = values.map(inferType);

  checkTypes(actualTypes, expectedTypes);

  var typeVariableValues = extractTypeVariableValues(values, actualTypes, expectedTypes);

  checkTypeVariableConsistency(typeVariableValues);
  checkConstraints(typeClasses, constraints, typeVariableValues);
}

// returns { a: [ 'a', ['a'] ], b: [ 'b' ] }
// for arrays, only returns the first value.
function extractTypeVariableValues(values, actualTypes, expectedTypes) {

  return zip(values, actualTypes, expectedTypes).reduce(function (typeVariableValues, _ref6) {
    var _ref7 = _slicedToArray(_ref6, 3),
        v = _ref7[0],
        at = _ref7[1],
        et = _ref7[2];

    if (!isTypeVariable(et)) return typeVariableValues;

    typeVariableValues[et.baseType] = typeVariableValues[et.baseType] || [];

    // If baseType is undefined, it's type couldn't be infered, so we can't
    // determine what value would be there in the array.
    if (et.depth > at.depth && at.baseType === 'undefined') {
      return typeVariableValues;
    }

    // todo: get infered types as parameter to avoid calculating it again.

    // [[a]] can't match [string] (however, [a] can match [[string]]).
    if (et.depth > at.depth) {
      throw new TypeError('Type list doesn\'t match actual values. Bad type depth:\n          expected ' + et + ', got ' + v + '.');
    }

    typeVariableValues[et.baseType].push(range(et.depth).reduce(function (v) {
      return v[0];
    }, v));

    return typeVariableValues;
  }, {});
}

function checkTypes(actualTypes, expectedTypes) {
  if (actualTypes.length !== expectedTypes.length) {
    throw new TypeError('Type list doesn\'t match actual values. Bad type count:\n      expected ' + expectedTypes.length + ', got ' + actualTypes.length);
  }

  zip(actualTypes, expectedTypes).forEach(function (_ref8) {
    var _ref9 = _slicedToArray(_ref8, 2),
        at = _ref9[0],
        et = _ref9[1];

    // Type variables are checked in checkTypeVariableConsistency().
    if (isTypeVariable(et)) return;

    if (!compareTypes(at, et)) {
      throw new TypeError('Type list doesn\'t match actual values. Wrong types:\n          expected ' + et + ', got ' + at);
    }
  });
}

// Checks all types of each type variable instance are the same.
// eg: 'a -> [a]' matches 'a'->['b'], but not 'a'->[1]
function checkTypeVariableConsistency(typeVariableValues) {

  Object.keys(typeVariableValues).map(function (k) {
    return [k, typeVariableValues[k]];
  }).map(function (_ref10) {
    var _ref11 = _slicedToArray(_ref10, 2),
        typeVariableName = _ref11[0],
        typeVariableValues = _ref11[1];

    return [typeVariableName, typeVariableValues.map(inferType)];
  }).forEach(function (_ref12) {
    var _ref13 = _slicedToArray(_ref12, 2),
        typeVariableName = _ref13[0],
        typeVariableTypes = _ref13[1];

    if (!isArrayHomogeneous(typeVariableTypes, compareTypes)) {
      throw new TypeError('Inconsistent type variable \'' + typeVariableName + '\':\n          expected homogenous types, got ' + typeVariableTypes + '.');
    }
  });
}

// [['a']] -> { depth: 2, baseType: 'string' }
// 1 -> { depth: 0, baseType: 'number' }
// [] -> { depth: 1, baseType: 'undefined' }
function inferType(value) {
  if (!Array.isArray(value)) {
    return createType(typeof value === 'undefined' ? 'undefined' : _typeof(value));
  }

  // Corner case: array is empty; can't infer it's type.
  // 'undefined' is used as wildcard.
  if (!value.length) return createType('undefined', 1);

  var typeOfFirstElement = inferType(value[0]);

  var typeOfElements = value.map(inferType);
  if (!isArrayHomogeneous(typeOfElements, compareTypes)) {
    // Treat Unhomogeneous arrays as objects.
    return createType('object');
  }

  return createType(typeOfFirstElement.baseType, typeOfFirstElement.depth + 1);
}

function compareTypes(a, b) {
  return a.baseType === b.baseType && a.depth === b.depth
  // Corner case: array was empty; couldn't infer type, so it matches anything.
  || a.baseType === 'undefined' && a.depth <= b.depth || b.baseType === 'undefined' && b.depth <= a.depth;
}

function isTypeVariable(type) {
  return !CONCRETE_TYPES.includes(type.baseType);
}

// typeClasses = { Num: fn(x), Ord: fn(x), ... }
// constraints = { a: [Num, Ord], b: [Num], ... }
// typeVariableValues = { a: ['a', 'b', 'c'], b: [1], c: [ [1, 2], [] ], ... }
function checkConstraints(typeClasses, constraints, typeVariableValues) {

  var variableNames = Object.keys(typeVariableValues);

  // Join all objects into a flat array of tuples like:
  // [ variable name, class name, predicate, value ]
  // Then, check whether all values pass their predicate.
  variableNames
  // Add type class names. (reduce is used as flatmap)
  .reduce(function (output, vn) {
    var classNames = constraints[vn] || [];
    return output.concat(classNames.map(function (cn) {
      return [vn, cn];
    }));
  }, [])
  // Filter out type variables without constraints.
  .filter(function (x) {
    return x.length;
  })
  // Add predicates.
  .map(function (_ref14) {
    var _ref15 = _slicedToArray(_ref14, 2),
        vn = _ref15[0],
        cn = _ref15[1];

    return [vn, cn, typeClasses[cn]];
  })
  // Filter out type classes without predicates.
  .filter(function (_ref16) {
    var _ref17 = _slicedToArray(_ref16, 3),
        p = _ref17[2];

    return p;
  })
  // Add values. (reduce is used as flatmap)
  .reduce(function (output, _ref18) {
    var _ref19 = _slicedToArray(_ref18, 3),
        vn = _ref19[0],
        cn = _ref19[1],
        p = _ref19[2];

    var values = typeVariableValues[vn];
    return output.concat(values.map(function (v) {
      return [vn, cn, p, v];
    }));
  }, [])
  // Actually check predicates pass.
  .forEach(function (_ref20) {
    var _ref21 = _slicedToArray(_ref20, 4),
        vn = _ref21[0],
        cn = _ref21[1],
        p = _ref21[2],
        v = _ref21[3];

    if (!p(v)) {
      throw new TypeError('Unmet class constraint \'' + cn + '\', on type variable \'' + vn + '\'');
    }
  });
}

function zip(a) {
  for (var _len2 = arguments.length, bs = Array(_len2 > 1 ? _len2 - 1 : 0), _key2 = 1; _key2 < _len2; _key2++) {
    bs[_key2 - 1] = arguments[_key2];
  }

  if (bs.some(function (b) {
    return a.length !== b.length;
  })) {
    throw Error('Can\'t zip arrays of different lengths');
  }

  return a.map(function (e, i) {
    return [e].concat(bs.map(function (b) {
      return b[i];
    }));
  });
}

function range(n) {
  return Array.apply(undefined, Array(n)).map(function (_, i) {
    return i;
  });
}

function isArrayHomogeneous(array, compare) {
  return array.every(function (x) {
    return compare(x, array[0]);
  });
}

function createType(baseType) {
  var depth = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : 0;

  return {
    baseType: baseType,
    depth: depth,
    toString: function toString() {
      var depth = this.depth,
          baseType = this.baseType;

      var leftBrackets = range(depth).map(function () {
        return '[';
      });
      var rightBrackets = range(depth).map(function () {
        return ']';
      });
      return '' + leftBrackets + baseType + rightBrackets;
    }
  };
}